"""Unit test for Sensor objects."""

from unittest.mock import AsyncMock, Mock, call

from xknx import XKNX
from xknx.core import XknxConnectionState
from xknx.devices import ExposeSensor
from xknx.dpt import DPTArray, DPTBinary
from xknx.telegram import GroupAddress, Telegram, TelegramDirection
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite

from ..conftest import EventLoopClockAdvancer


class TestExposeSensor:
    """Test class for Sensor objects."""

    #
    # STR FUNCTIONS
    #
    async def test_str_binary(self) -> None:
        """Test resolve state with binary sensor."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", group_address="1/2/3", value_type="binary"
        )
        expose_sensor.process(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(value=DPTBinary(1)),
            )
        )

        assert expose_sensor.resolve_state() is True
        assert expose_sensor.unit_of_measurement() is None

    async def test_str_percent(self) -> None:
        """Test resolve state with percent sensor."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", group_address="1/2/3", value_type="percent"
        )

        expose_sensor.process(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(DPTArray((0x40,))),
            )
        )

        assert expose_sensor.resolve_state() == 25
        assert expose_sensor.unit_of_measurement() == "%"

    async def test_str_temperature(self) -> None:
        """Test resolve state with temperature sensor."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", group_address="1/2/3", value_type="temperature"
        )

        expose_sensor.process(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(DPTArray((0x0C, 0x1A))),
            )
        )

        assert expose_sensor.resolve_state() == 21.0
        assert expose_sensor.unit_of_measurement() == "°C"

    #
    # TEST SET
    #
    async def test_set_binary(self) -> None:
        """Test set with binary sensor."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", group_address="1/2/3", value_type="binary"
        )
        await expose_sensor.set(False)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

    async def test_set_percent(self) -> None:
        """Test set with percent sensor."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", group_address="1/2/3", value_type="percent"
        )
        await expose_sensor.set(75)
        assert xknx.telegrams.qsize() == 1

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0xBF,))),
        )

    async def test_set_temperature(self) -> None:
        """Test set with temperature sensor."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", group_address="1/2/3", value_type="temperature"
        )
        await expose_sensor.set(21.0)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x0C, 0x1A))),
        )

    #
    # TEST PROCESS (GROUP READ)
    #
    async def test_process_binary(self) -> None:
        """Test reading binary expose sensor from bus."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", value_type="binary", group_address="1/2/3"
        )
        expose_sensor.sensor_value.value = True

        telegram = Telegram(GroupAddress("1/2/3"), payload=GroupValueRead())
        expose_sensor.process(telegram)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(DPTBinary(True)),
        )

    async def test_process_percent(self) -> None:
        """Test reading percent expose sensor from bus."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", value_type="percent", group_address="1/2/3"
        )

        expose_sensor.process(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(DPTArray((0x40,))),
            )
        )

        telegram = Telegram(GroupAddress("1/2/3"), payload=GroupValueRead())
        expose_sensor.process(telegram)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(DPTArray((0x40,))),
        )

    async def test_process_temperature(self) -> None:
        """Test reading temperature expose sensor from bus."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", value_type="temperature", group_address="1/2/3"
        )
        expose_sensor.sensor_value.value = 21.0

        telegram = Telegram(GroupAddress("1/2/3"), payload=GroupValueRead())
        expose_sensor.process(telegram)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(DPTArray((0x0C, 0x1A))),
        )

    async def test_process_no_respond_to_read(self) -> None:
        """Test expose sensor with respond_to_read set to False."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx,
            "TestSensor",
            value_type="temperature",
            group_address="1/2/3",
            respond_to_read=False,
        )
        expose_sensor.sensor_value.value = 21.0

        telegram = Telegram(GroupAddress("1/2/3"), payload=GroupValueRead())
        expose_sensor.process(telegram)
        assert xknx.telegrams.qsize() == 0

    #
    # HAS GROUP ADDRESS
    #
    def test_has_group_address(self) -> None:
        """Test expose sensor has group address."""
        xknx = XKNX()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", value_type="temperature", group_address="1/2/3"
        )
        assert expose_sensor.has_group_address(GroupAddress("1/2/3"))
        assert not expose_sensor.has_group_address(GroupAddress("1/2/4"))

    #
    # PROCESS CALLBACK
    #
    async def test_process_callback(self) -> None:
        """Test setting value. Test if callback is called."""

        xknx = XKNX()
        after_update_callback = Mock()
        expose_sensor = ExposeSensor(
            xknx, "TestSensor", group_address="1/2/3", value_type="temperature"
        )
        expose_sensor.register_device_updated_cb(after_update_callback)
        xknx.devices.async_add(expose_sensor)

        await expose_sensor.set(21.0)
        xknx.devices.process(xknx.telegrams.get_nowait())
        after_update_callback.assert_called_with(expose_sensor)

    #
    # TEST COOLDOWN
    #
    async def test_cooldown(self, time_travel: EventLoopClockAdvancer) -> None:
        """Test cooldown."""
        xknx = XKNX()
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        xknx.cemi_handler = AsyncMock()
        await xknx.telegram_queue.start()

        expose_sensor_cd = ExposeSensor(
            xknx,
            "TestSensor",
            group_address="1/2/3",
            value_type="temperature",
            cooldown=10,
        )
        xknx.devices.async_add(expose_sensor_cd)
        expose_sensor_no_cd = ExposeSensor(
            xknx,
            "TestSensor",
            group_address="1/2/4",
            value_type="temperature",
        )
        xknx.devices.async_add(expose_sensor_no_cd)

        await expose_sensor_cd.set(21.0)
        await expose_sensor_no_cd.set(21.0)
        await time_travel(0)
        assert xknx.cemi_handler.send_telegram.call_args_list == [
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=GroupValueWrite(DPTArray((0x0C, 0x1A))),
                )
            ),
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/4"),
                    payload=GroupValueWrite(DPTArray((0x0C, 0x1A))),
                )
            ),
        ]
        xknx.cemi_handler.send_telegram.reset_mock()

        # don't send telegram with same payload twice if cooldown is active
        await expose_sensor_cd.set(21.0)
        await expose_sensor_no_cd.set(21.0)
        await time_travel(0)
        assert xknx.cemi_handler.send_telegram.call_args_list == [
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/4"),
                    payload=GroupValueWrite(DPTArray((0x0C, 0x1A))),
                )
            ),
        ]

        xknx.cemi_handler.send_telegram.reset_mock()

        await time_travel(10)
        assert xknx.telegrams.qsize() == 0
        xknx.cemi_handler.send_telegram.assert_not_called()

        # different payload after cooldown
        await expose_sensor_cd.set(10.0)
        await expose_sensor_no_cd.set(10.0)
        await time_travel(0)
        assert xknx.cemi_handler.send_telegram.call_args_list == [
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=GroupValueWrite(DPTArray((0x03, 0xE8))),
                )
            ),
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/4"),
                    payload=GroupValueWrite(DPTArray((0x03, 0xE8))),
                )
            ),
        ]
        xknx.cemi_handler.send_telegram.reset_mock()
        # different payload in cooldown (payload of 3.111 equals 3.11)
        await expose_sensor_cd.set(3.111)
        await expose_sensor_no_cd.set(3.111)
        await time_travel(0)
        assert expose_sensor_cd._cooldown_task.done() is False
        assert xknx.cemi_handler.send_telegram.call_args_list == [
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/4"),
                    payload=GroupValueWrite(DPTArray((0x01, 0x37))),
                )
            ),
        ]
        xknx.cemi_handler.send_telegram.reset_mock()
        assert expose_sensor_cd.sensor_value.value == 10
        await time_travel(10)
        assert xknx.cemi_handler.send_telegram.call_args_list == [
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=GroupValueWrite(DPTArray((0x01, 0x37))),
                )
            ),
        ]
        xknx.cemi_handler.send_telegram.reset_mock()
        assert expose_sensor_cd.resolve_state() == 3.11
        await time_travel(10)
        assert expose_sensor_cd._cooldown_task.done()  # done after no change
        xknx.cemi_handler.send_telegram.assert_not_called()

        # reading unsent value
        await expose_sensor_cd.set(10)
        #   first send new value
        await time_travel(0)
        assert xknx.cemi_handler.send_telegram.call_args_list == [
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=GroupValueWrite(DPTArray((0x03, 0xE8))),
                )
            ),
        ]
        xknx.cemi_handler.send_telegram.reset_mock()
        #   in cooldown with a new value - receiving a read request
        await expose_sensor_cd.set(21)
        expose_sensor_cd.process(
            Telegram(GroupAddress("1/2/3"), payload=GroupValueRead())
        )
        await time_travel(0)
        assert xknx.cemi_handler.send_telegram.call_args_list == [
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=GroupValueResponse(DPTArray((0x0C, 0x1A))),
                )
            ),
        ]
        xknx.cemi_handler.send_telegram.reset_mock()
        #   after cooldown - new value not sent again (already in GroupValueResponse)
        await time_travel(10)
        xknx.cemi_handler.send_telegram.assert_not_called()

        # no connection blocks cooldown to avoid queueing unsent telegrams
        await expose_sensor_cd.set(20)
        await time_travel(1)
        xknx.cemi_handler.send_telegram.assert_called_once()
        xknx.cemi_handler.send_telegram.reset_mock()
        await expose_sensor_cd.set(30)
        await time_travel(5)
        xknx.connection_manager.connection_state_changed(
            XknxConnectionState.DISCONNECTED
        )
        await time_travel(5)
        await time_travel(10)
        xknx.cemi_handler.send_telegram.assert_not_called()
        assert not expose_sensor_cd._cooldown_task.done()
        await expose_sensor_cd.set(40)
        await time_travel(1)
        await expose_sensor_cd.set(50)
        await time_travel(1)
        await expose_sensor_cd.set(10)
        await time_travel(1)
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        await time_travel(0)
        assert xknx.cemi_handler.send_telegram.call_args_list == [
            call(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=GroupValueWrite(DPTArray((0x03, 0xE8))),
                )
            ),
        ]

        await xknx.stop()

    async def test_cooldown_read_request(
        self, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test cooldown for read requests."""
        xknx = XKNX()
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        xknx.cemi_handler = AsyncMock()
        await xknx.telegram_queue.start()

        expose_sensor_cd = ExposeSensor(
            xknx,
            "TestSensor",
            group_address="1/2/3",
            value_type="temperature",
            cooldown=10,
        )
        xknx.devices.async_add(expose_sensor_cd)
        expose_sensor_no_cd = ExposeSensor(
            xknx,
            "TestSensor",
            group_address="1/2/4",
            value_type="temperature",
        )
        xknx.devices.async_add(expose_sensor_no_cd)
        # reading during cooldown sends immediately, response starts cooldown
        await expose_sensor_cd.set(15)
        await time_travel(0)  # 0 after GroupValueWrite
        assert expose_sensor_cd._cooldown_task.done() is False
        xknx.cemi_handler.send_telegram.assert_called_once()
        xknx.cemi_handler.send_telegram.reset_mock()
        await time_travel(5)  # 5 after GroupValueWrite
        #   in cooldown - read request
        expose_sensor_cd.process(
            Telegram(GroupAddress("1/2/3"), payload=GroupValueRead())
        )
        await time_travel(0)  # 5 after GroupValueWrite (0 after GroupValueResponse)
        xknx.cemi_handler.send_telegram.assert_called_once()
        xknx.cemi_handler.send_telegram.reset_mock()
        assert expose_sensor_cd._cooldown_task.done() is False
        await time_travel(7)  # 12 after GroupValueWrite, 7 after GroupValueResponse
        assert expose_sensor_cd._cooldown_task.done() is False
        await time_travel(3)  # 15 after GroupValueWrite, 10 after GroupValueResponse
        assert expose_sensor_cd._cooldown_task.done() is True

        await xknx.stop()

    async def test_set_same_payload(self, time_travel: EventLoopClockAdvancer) -> None:
        """Test expose sensor with similar payloads in cooldown."""
        xknx = XKNX()
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        expose_sensor = ExposeSensor(
            xknx,
            "TestSensor",
            group_address="1/2/3",
            value_type="temperature",
            cooldown=10,
        )

        telegram_21_degree = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x0C, 0x1A))),
        )

        await expose_sensor.set(21.000123, skip_unchanged=True)  # 21.0 when decoded
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == telegram_21_degree

        expose_sensor.process(telegram)
        assert expose_sensor.resolve_state() == 21.0

        # in cooldown, similar payload, no telegram sent
        assert expose_sensor._cooldown_task.done() is False
        await expose_sensor.set(21.000567, skip_unchanged=True)  # 21.0 when decoded
        await time_travel(10)
        assert xknx.telegrams.empty()

        # outside cooldown, similar payload, no telegram sent
        await time_travel(10)
        assert expose_sensor._cooldown_task.done() is True
        await expose_sensor.set(21.00089, skip_unchanged=True)
        assert xknx.telegrams.empty()

        # no cooldown for non-sent telegram, new payload, telegram sent immediately
        assert expose_sensor._cooldown_task.done() is True
        await expose_sensor.set(22.0, skip_unchanged=True)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram.payload.value == DPTArray((0x0C, 0x4C))
        expose_sensor.process(telegram)

        # in cooldown getting new payload and then previous payload again - shall skip sending
        assert expose_sensor._cooldown_task.done() is False
        await expose_sensor.set(21.0, skip_unchanged=True)
        await expose_sensor.set(22.0, skip_unchanged=True)
        await time_travel(10)
        assert xknx.telegrams.empty()

    async def test_periodic_send(
        self,
        xknx_no_interface: XKNX,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test periodic send of expose sensor."""
        xknx = xknx_no_interface
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        xknx.cemi_handler = AsyncMock()
        expose_sensor = ExposeSensor(
            xknx,
            "TestSensor",
            group_address="1/2/3",
            value_type="switch",
            cooldown=2,
            periodic_send=10,
        )
        xknx.devices.async_add(expose_sensor)
        await xknx.start()

        test_telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(True)),
        )

        # uninitialized doesn't send anything
        await time_travel(10)
        xknx.cemi_handler.send_telegram.assert_not_called()

        # set a value - send immediately
        await expose_sensor.set(True)
        await time_travel(0)
        xknx.cemi_handler.send_telegram.assert_called_once_with(test_telegram)
        xknx.cemi_handler.send_telegram.reset_mock()

        # no resend before periodic time
        await time_travel(9)
        xknx.cemi_handler.send_telegram.assert_not_called()

        # resend after periodic time
        await time_travel(1)
        xknx.cemi_handler.send_telegram.assert_called_once_with(test_telegram)
        xknx.cemi_handler.send_telegram.reset_mock()

        # incoming update doesn't reset periodic timer
        await time_travel(7)
        expose_sensor.process(
            Telegram(
                direction=TelegramDirection.INCOMING,
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(DPTBinary(True)),
            )
        )
        await time_travel(2)
        xknx.cemi_handler.send_telegram.assert_not_called()
        await time_travel(1)
        xknx.cemi_handler.send_telegram.assert_called_once_with(test_telegram)
        xknx.cemi_handler.send_telegram.reset_mock()

        # outgoing cooldown telegram resets periodic timer
        await time_travel(7)
        await expose_sensor.set(False)
        await expose_sensor.set(True)  # within cooldown
        await time_travel(0)
        xknx.cemi_handler.send_telegram.assert_called_once()  # with False, second is in cooldown
        xknx.cemi_handler.send_telegram.reset_mock()
        await time_travel(2)
        xknx.cemi_handler.send_telegram.assert_called_once_with(test_telegram)
        xknx.cemi_handler.send_telegram.reset_mock()
        #    after cooldown, periodic timer reset - wait full periodic time
        await time_travel(9)
        xknx.cemi_handler.send_telegram.assert_not_called()
        await time_travel(1)
        xknx.cemi_handler.send_telegram.assert_called_once_with(test_telegram)
        xknx.cemi_handler.send_telegram.reset_mock()
        #    periodic telegram triggers cooldown
        await expose_sensor.set(False)
        await time_travel(0)
        xknx.cemi_handler.send_telegram.assert_not_called()
        xknx.cemi_handler.send_telegram.reset_mock()
        await time_travel(2)
        xknx.cemi_handler.send_telegram.assert_called_once_with(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(DPTBinary(False)),
            )
        )
        xknx.cemi_handler.send_telegram.reset_mock()
        # don't try to send without connection
        xknx.connection_manager.connection_state_changed(
            XknxConnectionState.DISCONNECTED
        )
        await time_travel(10)
        xknx.cemi_handler.send_telegram.assert_not_called()

        await xknx.stop()



"""
KNXIPInterface manages KNX/IP Tunneling or Routing connections.

* It searches for available devices and connects with the corresponding connect method.
* It passes KNX telegrams from the network and
* provides callbacks after having received a telegram from the network.

"""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine
import logging
import threading
from typing import TYPE_CHECKING, Any, TypeVar

from xknx.cemi import CEMIFrame
from xknx.exceptions import (
    CommunicationError,
    InvalidSecureConfiguration,
    XKNXException,
)
from xknx.io import util
from xknx.secure.keyring import InterfaceType, Keyring, XMLInterface, load_keyring
from xknx.telegram import IndividualAddress

from .connection import ConnectionConfig, ConnectionType
from .const import DEFAULT_INDIVIDUAL_ADDRESS
from .gateway_scanner import GatewayDescriptor, GatewayScanner
from .routing import Routing, SecureRouting
from .self_description import request_description
from .tunnel import SecureTunnel, TCPTunnel, UDPTunnel, _Tunnel

if TYPE_CHECKING:
    import concurrent

    from xknx.xknx import XKNX

    from .interface import Interface

logger = logging.getLogger("xknx.log")

T = TypeVar("T")  # pylint: disable=invalid-name


def knx_interface_factory(
    xknx: XKNX,
    connection_config: ConnectionConfig,
) -> KNXIPInterface:
    """Create KNX/IP interface from config."""
    if connection_config.threaded:
        return KNXIPInterfaceThreaded(xknx=xknx, connection_config=connection_config)
    return KNXIPInterface(xknx=xknx, connection_config=connection_config)


class KNXIPInterface:
    """Class for managing KNX/IP Tunneling or Routing connections."""

    __slots__ = ("_gateway_info", "_interface", "connection_config", "xknx")

    def __init__(
        self,
        xknx: XKNX,
        connection_config: ConnectionConfig | None = None,
    ) -> None:
        """Initialize KNXIPInterface class."""
        self.xknx = xknx
        self.connection_config = connection_config or ConnectionConfig()
        self._gateway_info: GatewayDescriptor | None = None
        self._interface: Interface | None = None

    async def start(self) -> None:
        """Start KNX/IP interface. Raise `CommunicationError` if connection fails."""
        await self._start()

    async def _start(self) -> None:
        """Start interface. Connecting KNX/IP device with the selected method."""
        if gateway_ip := self.connection_config.gateway_ip:
            gateway_ip = await util.validate_ip(gateway_ip, address_name="Gateway IP")
        if local_ip := self.connection_config.local_ip:
            local_ip = await util.validate_ip(local_ip, address_name="Local IP")
        keyring: Keyring | None = None
        if secure_config := self.connection_config.secure_config:
            if secure_config.keyring is not None:
                keyring = secure_config.keyring
            elif (
                secure_config.knxkeys_file_path is not None
                and secure_config.knxkeys_password is not None
            ):
                keyring = await load_keyring(
                    secure_config.knxkeys_file_path,
                    secure_config.knxkeys_password,
                )
        self.xknx.cemi_handler.data_secure_init(keyring=keyring)

        if self.connection_config.connection_type == ConnectionType.ROUTING:
            await self._start_routing(local_ip=local_ip)
        elif self.connection_config.connection_type == ConnectionType.ROUTING_SECURE:
            await self._start_secure_routing(local_ip=local_ip, keyring=keyring)
        elif (
            self.connection_config.connection_type == ConnectionType.TUNNELING
            and gateway_ip is not None
        ):
            await self._start_tunnelling_udp(
                gateway_ip=gateway_ip,
                gateway_port=self.connection_config.gateway_port,
                local_ip=local_ip,
            )
        elif (
            self.connection_config.connection_type == ConnectionType.TUNNELING_TCP
            and gateway_ip is not None
        ):
            await self._start_tunnelling_tcp(
                gateway_ip=gateway_ip,
                gateway_port=self.connection_config.gateway_port,
            )
        elif (
            self.connection_config.connection_type
            == ConnectionType.TUNNELING_TCP_SECURE
            and gateway_ip is not None
        ):
            await self._start_secure_tunnelling_tcp(
                gateway_ip=gateway_ip,
                gateway_port=self.connection_config.gateway_port,
                keyring=keyring,
            )
        else:
            await self._start_automatic(local_ip=local_ip, keyring=keyring)

    async def _start_automatic(
        self,
        local_ip: str | None,
        keyring: Keyring | None,
    ) -> None:
        """Start GatewayScanner and connect to the found device."""
        keyring_host_filter: set[IndividualAddress] = set()
        if keyring:
            if required_addr := self.connection_config.individual_address:
                _host_ia = keyring.get_tunnel_host_by_interface(
                    tunnelling_slot=required_addr
                )
                if _host_ia is None:
                    raise InvalidSecureConfiguration(
                        f"No host for required address {required_addr} found in keyring file."
                    )
                keyring_host_filter.add(_host_ia)
            else:
                keyring_host_filter.update(
                    interface.host
                    for interface in keyring.interfaces
                    if interface.host is not None
                    and interface.type is InterfaceType.TUNNELING
                )
        async for gateway in GatewayScanner(
            self.xknx,
            local_ip=local_ip,
            scan_filter=self.connection_config.scan_filter,
        ).async_scan():
            if (
                keyring_host_filter
                and gateway.individual_address not in keyring_host_filter
            ):
                logger.debug("Skipping %s. No match in keyring file", gateway)
                continue
            try:
                if gateway.supports_tunnelling_tcp:
                    if gateway.tunnelling_requires_secure:
                        await self._start_secure_tunnelling_tcp(
                            gateway_ip=gateway.ip_addr,
                            gateway_port=gateway.port,
                            gateway_descriptor=gateway,
                            keyring=keyring,
                        )
                    else:
                        await self._start_tunnelling_tcp(
                            gateway_ip=gateway.ip_addr,
                            gateway_port=gateway.port,
                        )
                elif (
                    gateway.supports_tunnelling
                    and not gateway.tunnelling_requires_secure
                ):
                    await self._start_tunnelling_udp(
                        gateway_ip=gateway.ip_addr,
                        gateway_port=gateway.port,
                        local_ip=local_ip,
                    )
                elif gateway.supports_routing and not gateway.routing_requires_secure:
                    await self._start_routing(local_ip=local_ip)
            except CommunicationError as ex:
                logger.debug("Skipping %s. Could not connect: %s", gateway, ex)
                continue
            except InvalidSecureConfiguration as ex:
                logger.debug(
                    "Skipping %s. Invalid secure configuration: %s", gateway, ex
                )
                continue
            else:
                self._gateway_info = gateway
                break
        else:
            raise CommunicationError(
                f"No usable KNX/IP device found{' in keyring file' if keyring_host_filter else ''}."
            )

    async def _start_tunnelling_tcp(
        self,
        gateway_ip: str,
        gateway_port: int,
    ) -> None:
        """Start KNX/IP TCP tunnel."""
        tunnel_address = self.connection_config.individual_address

        logger.debug(
            "Starting tunnel to %s:%s over TCP%s",
            gateway_ip,
            gateway_port,
            f" requesting individual address {tunnel_address}"
            if tunnel_address
            else "",
        )
        self._interface = TCPTunnel(
            self.xknx,
            gateway_ip=gateway_ip,
            gateway_port=gateway_port,
            individual_address=tunnel_address,
            cemi_received_callback=self.cemi_received,
            auto_reconnect=self.connection_config.auto_reconnect,
            auto_reconnect_wait=self.connection_config.auto_reconnect_wait,
        )
        await self._interface.connect()

    async def _start_secure_tunnelling_tcp(
        self,
        gateway_ip: str,
        gateway_port: int,
        gateway_descriptor: GatewayDescriptor | None = None,
        keyring: Keyring | None = None,
    ) -> None:
        """Start KNX/IP Secure TCP tunnel."""
        if (secure_config := self.connection_config.secure_config) is None:
            raise InvalidSecureConfiguration("SecureConfig missing in ConnectionConfig")

        if (
            secure_config.user_id is not None
            and secure_config.user_password is not None
        ):
            user_id = secure_config.user_id
            user_password = secure_config.user_password
            device_authentication_password = (
                secure_config.device_authentication_password
            )
        elif keyring is not None:
            _gateway = gateway_descriptor or await request_description(
                gateway_ip=gateway_ip, gateway_port=gateway_port
            )
            xml_interface = self._get_tunnel_interface_from_keyring(
                keyring=keyring,
                gateway_descriptor=_gateway,
                individual_address=self.connection_config.individual_address,
                config_user_id=secure_config.user_id,
            )
            if (
                xml_interface.user_id is None
                or xml_interface.decrypted_password is None
            ):
                raise InvalidSecureConfiguration(
                    f"No user_id or password found for tunnel {xml_interface.individual_address}"
                )
            user_id = xml_interface.user_id
            user_password = xml_interface.decrypted_password
            device_authentication_password = xml_interface.decrypted_authentication
        else:
            raise InvalidSecureConfiguration(
                "No `user_id` or `knxkeys_file_path` and password found in secure configuration"
            )

        logger.debug(
            "Starting secure tunnel to %s:%s over TCP",
            gateway_ip,
            gateway_port,
        )
        self._interface = SecureTunnel(
            self.xknx,
            gateway_ip=gateway_ip,
            gateway_port=gateway_port,
            auto_reconnect=self.connection_config.auto_reconnect,
            auto_reconnect_wait=self.connection_config.auto_reconnect_wait,
            user_id=user_id,
            user_password=user_password,
            device_authentication_password=device_authentication_password,
            cemi_received_callback=self.cemi_received,
        )
        await self._interface.connect()

    @staticmethod
    def _get_tunnel_interface_from_keyring(
        keyring: Keyring,
        gateway_descriptor: GatewayDescriptor,
        individual_address: IndividualAddress | None = None,
        config_user_id: int | None = None,
    ) -> XMLInterface:
        """
        Get tunnel interface from keyring.

        Precedence: configured individual address > configured user id > first free tunnel interface
        """
        if individual_address:
            if xml_interface := keyring.get_tunnel_interface_by_individual_address(
                individual_address
            ):
                return xml_interface
            raise InvalidSecureConfiguration(
                f"Interface with individual address {individual_address} not found in keyfile"
            )

        if not (host_ia := gateway_descriptor.individual_address):
            raise InvalidSecureConfiguration(
                "Gateway does not provide individual address"
            )
        if config_user_id:
            if xml_interface := keyring.get_tunnel_interface_by_host_and_user_id(
                host=host_ia, user_id=config_user_id
            ):
                return xml_interface
            raise InvalidSecureConfiguration(
                f"Interface of host {host_ia} with user_id {config_user_id} not found in keyfile"
            )

        _free_slots = [
            tunnel_ia
            for tunnel_ia, slot in gateway_descriptor.tunnelling_slots.items()
            if slot.usable and slot.free
        ]
        if not _free_slots:
            raise InvalidSecureConfiguration(
                f"No free tunnelling slots found on gateway {host_ia}"
            )
        for _tunnel_ia in _free_slots:
            if xml_interface := keyring.get_tunnel_interface_by_individual_address(
                tunnelling_slot=_tunnel_ia
            ):
                return xml_interface

        raise InvalidSecureConfiguration(
            "No credentials for any free tunnelling slot found in keyfile"
        )

    async def _start_tunnelling_udp(
        self,
        gateway_ip: str,
        gateway_port: int,
        local_ip: str | None,
    ) -> None:
        """Start KNX/IP UDP tunnel."""
        local_port = self.connection_config.local_port
        route_back = self.connection_config.route_back

        local_ip = local_ip or util.find_local_ip(gateway_ip=gateway_ip)
        if local_ip is None:
            local_ip = await util.get_default_local_ip(gateway_ip)
            if local_ip is None:
                raise XKNXException("No network interface found.")
            route_back = True
            logger.debug("Falling back to default interface and enabling route back.")

        logger.debug(
            "Starting tunnel from %s:%s to %s:%s",
            local_ip,
            local_port,
            gateway_ip,
            gateway_port,
        )
        self._interface = UDPTunnel(
            self.xknx,
            gateway_ip=gateway_ip,
            gateway_port=gateway_port,
            local_ip=local_ip,
            local_port=local_port,
            route_back=route_back,
            cemi_received_callback=self.cemi_received,
            auto_reconnect=self.connection_config.auto_reconnect,
            auto_reconnect_wait=self.connection_config.auto_reconnect_wait,
        )
        await self._interface.connect()

    async def _start_routing(self, local_ip: str | None) -> None:
        """Start KNX/IP Routing."""
        multicast_group = self.connection_config.multicast_group
        multicast_port = self.connection_config.multicast_port

        local_ip = local_ip or await util.get_default_local_ip()
        if local_ip is None:
            raise XKNXException("No network interface found.")

        individual_address = (
            self.connection_config.individual_address or DEFAULT_INDIVIDUAL_ADDRESS
        )

        logger.debug(
            "Starting Routing from %s as %s via %s:%s",
            local_ip,
            individual_address,
            multicast_group,
            multicast_port,
        )
        self._interface = Routing(
            self.xknx,
            individual_address=individual_address,
            cemi_received_callback=self.cemi_received,
            local_ip=local_ip,
            multicast_group=multicast_group,
            multicast_port=multicast_port,
        )
        await self._interface.connect()

    async def _start_secure_routing(
        self,
        local_ip: str | None,
        keyring: Keyring | None = None,
    ) -> None:
        """Start KNX/IP Routing."""
        if self.connection_config.secure_config is None:
            raise InvalidSecureConfiguration("SecureConfig missing in ConnectionConfig")
        backbone_key = self.connection_config.secure_config.backbone_key
        latency_ms = self.connection_config.secure_config.latency_ms
        multicast_group = self.connection_config.multicast_group
        multicast_port = self.connection_config.multicast_port

        if keyring is not None and keyring.backbone is not None:
            # default to manually configured values
            backbone_key = backbone_key or keyring.backbone.decrypted_key
            latency_ms = latency_ms or keyring.backbone.latency
            if keyring.backbone.multicast_address:
                multicast_group = keyring.backbone.multicast_address
        if not backbone_key:
            raise InvalidSecureConfiguration(
                "No backbone key found in secure configuration"
            )

        local_ip = local_ip or await util.get_default_local_ip()
        if local_ip is None:
            raise XKNXException("No network interface found.")

        individual_address = (
            self.connection_config.individual_address or DEFAULT_INDIVIDUAL_ADDRESS
        )

        logger.debug(
            "Starting Secure Routing from %s as %s via %s:%s",
            local_ip,
            individual_address,
            multicast_group,
            multicast_port,
        )
        self._interface = SecureRouting(
            self.xknx,
            individual_address=individual_address,
            cemi_received_callback=self.cemi_received,
            local_ip=local_ip,
            backbone_key=backbone_key,
            latency_ms=latency_ms,
            multicast_group=multicast_group,
            multicast_port=multicast_port,
        )
        await self._interface.connect()

    async def stop(self) -> None:
        """Stop connected interfae (either Tunneling or Routing)."""
        if self._interface is not None:
            await self._interface.disconnect()
            self._interface = None

    def cemi_received(self, raw_cemi: bytes) -> None:
        """Pass raw CEMIFrame data to CEMIHandler. Callback for having received CEMIFrames."""
        self.xknx.cemi_handler.handle_raw_cemi(raw_cemi)

    async def send_cemi(self, cemi: CEMIFrame) -> None:
        """Send CEMIFrame to connected device (either Tunneling or Routing)."""
        # to ease converting L_Data.req CEMI frames to L_Data.ind and local confirmation
        # in routing we pass `CEMIFrame` from the CEMIHandler here, instead of raw bytes
        if self._interface is None:
            raise CommunicationError("KNX/IP interface not connected")
        return await self._interface.send_cemi(cemi)

    async def gateway_info(self) -> GatewayDescriptor | None:
        """Get gateway descriptor from interface."""
        if self._gateway_info is not None:
            return self._gateway_info
        if isinstance(self._interface, _Tunnel):
            return await self._interface.request_description()
        return None


class KNXIPInterfaceThreaded(KNXIPInterface):
    """Class for managing KNX/IP Tunneling or Routing connections."""

    __slots__ = ("_connection_thread", "_main_loop", "_thread_loop")

    def __init__(
        self,
        xknx: XKNX,
        connection_config: ConnectionConfig | None = None,
    ) -> None:
        """Initialize KNXIPInterface class."""
        super().__init__(xknx, connection_config)
        self._main_loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        self._connection_thread: threading.Thread | None = None
        self._thread_loop: asyncio.AbstractEventLoop | None = None

    def _init_connection_thread(self) -> None:
        """Start KNX/IP interface in its own thread."""
        loop_loaded = threading.Event()
        self._connection_thread = threading.Thread(
            target=self._init_connection_loop,
            args=[loop_loaded],
            name="KNX Interface",
            daemon=True,
        )
        self._connection_thread.start()
        loop_loaded.wait()  # wait for the thread to initialize its loop

    def _init_connection_loop(self, loop_loaded: threading.Event) -> None:
        """Start KNX/IP interface in its own thread."""
        self._thread_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._thread_loop)
        loop_loaded.set()
        self._thread_loop.run_forever()

    async def _await_from_connection_thread(self, coro: Coroutine[Any, Any, T]) -> T:
        """Await coroutine in different thread."""
        if self._thread_loop is None:
            raise CommunicationError("KNX connection thread not initialized.")

        fut = asyncio.run_coroutine_threadsafe(coro, self._thread_loop)
        finished = threading.Event()

        def fut_finished_cb(_: concurrent.futures.Future[T]) -> None:
            """Fire threading.Event when the future is finished."""
            finished.set()

        fut.add_done_callback(fut_finished_cb)
        # wait on that event in an executor, yielding control to _main_loop
        await self._main_loop.run_in_executor(None, finished.wait)
        return fut.result()

    async def start(self) -> None:
        """Start KNX/IP interface."""
        if self._connection_thread is not None or self._thread_loop is not None:
            raise CommunicationError("KNX threaded interface already initialized.")
        await self._main_loop.run_in_executor(None, self._init_connection_thread)
        try:
            return await self._await_from_connection_thread(self._start())
        except CommunicationError:
            await self.stop()
            raise

    async def stop(self) -> None:
        """Stop connected interface (either Tunneling or Routing)."""
        if self._interface is not None:
            await self._await_from_connection_thread(self._interface.disconnect())
            self._interface = None
        if self._thread_loop is not None:
            self._thread_loop.call_soon_threadsafe(self._thread_loop.stop)
            self._thread_loop = None
        if self._connection_thread is not None:
            self._connection_thread.join()
            self._connection_thread = None

    def cemi_received(self, raw_cemi: bytes) -> None:
        """Pass CEMIFrame to CEMIHandler. Callback for having received CEMIFrames."""
        self._main_loop.call_soon_threadsafe(super().cemi_received, raw_cemi)

    async def send_cemi(self, cemi: CEMIFrame) -> None:
        """Send CEMIFrame to connected device (either Tunneling or Routing)."""
        if self._interface is None:
            raise CommunicationError("KNX/IP interface not connected")

        return await self._await_from_connection_thread(self._interface.send_cemi(cemi))

    async def gateway_info(self) -> GatewayDescriptor | None:
        """Get gateway descriptor from interface."""
        if self._gateway_info is not None:
            return self._gateway_info
        if isinstance(self._interface, _Tunnel):
            return await self._await_from_connection_thread(
                self._interface.request_description()
            )
        return None



"""
Module for Serialization and Deserialization of KNX Timer Notify.

This frame shall be sent during secure KNXnet/IP multicast group communication
to keep the multicast group member's timer values synchronized.
"""

from __future__ import annotations

from typing import Final

from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBody
from .knxip_enum import KNXIPServiceType


class TimerNotify(KNXIPBody):
    """Representation of a KNX Timer Notify."""

    SERVICE_TYPE = KNXIPServiceType.TIMER_NOTIFY
    LENGTH: Final = 30

    def __init__(
        self,
        timer_value: int = 0,
        serial_number: bytes = bytes(6),
        message_tag: bytes = bytes(2),
        message_authentication_code: bytes = bytes(16),
    ) -> None:
        """Initialize TimerNotify object."""
        self.timer_value = timer_value
        self.serial_number = serial_number
        self.message_tag = message_tag
        self.message_authentication_code = message_authentication_code

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return TimerNotify.LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) != TimerNotify.LENGTH:
            raise CouldNotParseKNXIP("TimerNotify has wrong length")
        self.timer_value = int.from_bytes(raw[:6], "big")
        self.serial_number = raw[6:12]
        self.message_tag = raw[12:14]
        self.message_authentication_code = raw[14:]
        return TimerNotify.LENGTH

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            self.timer_value.to_bytes(6, "big")
            + self.serial_number
            + self.message_tag
            + self.message_authentication_code
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<TimerNotify "
            f'timer_value="{self.timer_value}" '
            f'serial_number="{self.serial_number.hex()}" '
            f'message_tag="{self.message_tag.hex()}" '
            f'message_authentication_code="{self.message_authentication_code.hex()}" />'
        )



"""Unit test for Notification objects."""

from unittest.mock import Mock, patch

from xknx import XKNX
from xknx.devices import Notification
from xknx.dpt import DPTArray, DPTBinary, DPTString
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite


class TestNotification:
    """Test class for Notification object."""

    #
    # SYNC
    #
    async def test_sync_state(self) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        notification = Notification(
            xknx, "Warning", group_address="1/2/3", group_address_state="1/2/4"
        )
        await notification.sync()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/4"), payload=GroupValueRead()
        )

    #
    # TEST PROCESS
    #
    async def test_process(self) -> None:
        """Test process telegram with notification. Test if device was updated."""
        xknx = XKNX()
        notification = Notification(xknx, "Warning", group_address="1/2/3")
        telegram_set = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTString().to_knx("Ein Prosit!")),
        )
        notification.process(telegram_set)
        assert notification.message == "Ein Prosit!"

        telegram_unset = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTString().to_knx("")),
        )
        notification.process(telegram_unset)
        assert notification.message == ""

    async def test_process_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback was called."""

        xknx = XKNX()
        notification = Notification(xknx, "Warning", group_address="1/2/3")
        after_update_callback = Mock()
        notification.register_device_updated_cb(after_update_callback)

        telegram_set = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTString().to_knx("Ein Prosit!")),
        )
        notification.process(telegram_set)
        after_update_callback.assert_called_with(notification)

    async def test_process_payload_invalid_length(self) -> None:
        """Test process wrong telegram (wrong payload length)."""
        xknx = XKNX()
        after_update_callback = Mock()
        notification = Notification(
            xknx,
            "Warning",
            group_address="1/2/3",
            device_updated_cb=after_update_callback,
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((23, 24))),
        )
        with patch("logging.Logger.warning") as log_mock:
            notification.process(telegram)
            log_mock.assert_called_once()
            after_update_callback.assert_not_called()

    async def test_process_wrong_payload(self) -> None:
        """Test process wrong telegram (wrong payload type)."""
        xknx = XKNX()
        after_update_callback = Mock()
        notification = Notification(
            xknx,
            "Warning",
            group_address="1/2/3",
            device_updated_cb=after_update_callback,
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        with patch("logging.Logger.warning") as log_mock:
            notification.process(telegram)
            log_mock.assert_called_once()
            after_update_callback.assert_not_called()

    #
    # TEST RESPOND
    #
    async def test_respond_to_read(self) -> None:
        """Test respond_to_read function."""
        xknx = XKNX()
        responding = Notification(
            xknx,
            "TestSensor1",
            group_address="1/1/1",
            respond_to_read=True,
            value_type="latin_1",
        )
        non_responding = Notification(
            xknx,
            "TestSensor2",
            group_address="1/1/1",
            respond_to_read=False,
            value_type="latin_1",
        )
        #  set initial payload of Notification
        responding.remote_value.value = "Halli Hallo"
        non_responding.remote_value.value = "Halli Hallo"

        read_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"), payload=GroupValueRead()
        )
        # verify no response when respond_to_read is False
        non_responding.process(read_telegram)
        assert xknx.telegrams.qsize() == 0

        # verify response when respond_to_read is True
        responding.process(read_telegram)
        assert xknx.telegrams.qsize() == 1
        response = xknx.telegrams.get_nowait()
        assert response == Telegram(
            destination_address=GroupAddress("1/1/1"),
            payload=GroupValueResponse(
                DPTArray(
                    (
                        0x48,
                        0x61,
                        0x6C,
                        0x6C,
                        0x69,
                        0x20,
                        0x48,
                        0x61,
                        0x6C,
                        0x6C,
                        0x6F,
                        0x0,
                        0x0,
                        0x0,
                    ),
                ),
            ),
        )

    #
    # TEST SET MESSAGE
    #
    async def test_set(self) -> None:
        """Test notificationing off notification."""
        xknx = XKNX()
        notification = Notification(xknx, "Warning", group_address="1/2/3")
        await notification.set("Ein Prosit!")
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTString().to_knx("Ein Prosit!")),
        )
        # test if message longer than 14 chars gets cropped
        await notification.set("This is too long.")

        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTString().to_knx("This is too lo")),
        )

    #
    # TEST has_group_address
    #
    def test_has_group_address(self) -> None:
        """Test has_group_address."""
        xknx = XKNX()
        notification = Notification(
            xknx, "Warning", group_address="1/2/3", group_address_state="1/2/4"
        )
        assert notification.has_group_address(GroupAddress("1/2/3"))
        assert notification.has_group_address(GroupAddress("1/2/4"))
        assert not notification.has_group_address(GroupAddress("2/2/2"))



"""Class for Data Secure Application layer service data units (ASDUs)."""

from __future__ import annotations

from enum import IntEnum

from xknx.exceptions import DataSecureError
from xknx.telegram.tpci import TPCI

from .security_primitives import (
    calculate_message_authentication_code_cbc,
    decrypt_ctr,
    encrypt_data_ctr,
)

# Secure APCI is 0x03F1 - in block_0 it is used split into 2 octets
_APCI_SEC_HIGH = 0x03
_APCI_SEC_LOW = 0xF1
# only Address Type (IA / GA) and Extended Frame format are used
B0_AT_FIELD_FLAGS_MASK = 0b10001111


def block_0(
    sequence_number: bytes,
    address_fields_raw: bytes,
    frame_flags: int,
    tpci_int: int,
    payload_length: int,
) -> bytes:
    """Return Block 0 for KNX Data Secure."""
    return (
        sequence_number
        + address_fields_raw
        + bytes(
            (
                0,
                frame_flags & B0_AT_FIELD_FLAGS_MASK,
                (tpci_int << 2) + _APCI_SEC_HIGH,
                _APCI_SEC_LOW,
                0,
                payload_length,
            )
        )
    )


def counter_0(
    sequence_number: bytes,
    address_fields_raw: bytes,
) -> bytes:
    """Return Block Counter 0 for KNX Data Secure."""
    return sequence_number + address_fields_raw + b"\x00\x00\x00\x00\x01\x00"


class SecurityAlgorithmIdentifier(IntEnum):
    """Enum representing the used security algorithm."""

    CCM_AUTHENTICATION = 0b000
    CCM_ENCRYPTION = 0b001


class SecurityALService(IntEnum):
    """Enum representing the used security application layer service."""

    S_A_DATA = 0b000
    S_A_SYNC_REQ = 0b001
    S_A_SYNC_RES = 0b011


class SecurityControlField:
    """Class for KNX Data Secure Security Control Field (SCF)."""

    __slots__ = ("algorithm", "service", "system_broadcast", "tool_access")

    def __init__(
        self,
        tool_access: bool,
        algorithm: SecurityAlgorithmIdentifier,
        system_broadcast: bool,
        service: SecurityALService,
    ) -> None:
        """Initialize SecurityControlField class."""
        self.tool_access = tool_access
        self.algorithm = algorithm
        self.system_broadcast = system_broadcast
        self.service = service

    @staticmethod
    def from_knx(raw: int) -> SecurityControlField:
        """Parse/deserialize from KNX raw data."""
        tool_access = bool(raw & 0b10000000)
        sai = SecurityAlgorithmIdentifier(raw >> 4 & 0b111)
        system_broadcast = bool(raw & 0b1000)
        s_al_service = SecurityALService(raw & 0b111)

        return SecurityControlField(
            tool_access=tool_access,
            algorithm=sai,
            system_broadcast=system_broadcast,
            service=s_al_service,
        )

    def to_knx(self) -> bytes:
        """Serialize to KNX raw data."""
        raw = 0
        raw |= self.tool_access << 7
        raw |= self.algorithm << 4
        raw |= self.system_broadcast << 3
        raw |= self.service
        return raw.to_bytes(1, "big")

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            f"<SecurityControlField tool_access={self.tool_access} "
            f"algorithm={self.algorithm.name} "
            f"system_broadcast={self.system_broadcast} "
            f"service={self.service.name} />"
        )


class SecureData:
    """Class for KNX Data Secure ASDU for S-A_Data-service."""

    __slots__ = ("message_authentication_code", "secured_apdu", "sequence_number_bytes")

    def __init__(
        self,
        sequence_number_bytes: bytes,
        secured_apdu: bytes,
        message_authentication_code: bytes,
    ) -> None:
        """Initialize SecureData class."""
        self.sequence_number_bytes = sequence_number_bytes
        self.secured_apdu = secured_apdu
        self.message_authentication_code = message_authentication_code

    def __len__(self) -> int:
        """Return length of KNX Data Secure ASDU."""
        return 10 + len(self.secured_apdu)  # 10 = 6 bytes sequence number + 4 bytes MAC

    @staticmethod
    def init_from_plain_apdu(
        key: bytes,
        apdu: bytes | bytearray,
        scf: SecurityControlField,
        sequence_number: int,
        address_fields_raw: bytes,
        frame_flags: int,
        tpci: TPCI,
    ) -> SecureData:
        """Serialize to KNX raw data."""
        sequence_number_bytes = sequence_number.to_bytes(6, "big")

        if scf.algorithm == SecurityAlgorithmIdentifier.CCM_AUTHENTICATION:
            mac = calculate_message_authentication_code_cbc(
                key=key,
                additional_data=scf.to_knx() + apdu,
                block_0=block_0(
                    sequence_number=sequence_number_bytes,
                    address_fields_raw=address_fields_raw,
                    frame_flags=frame_flags,
                    tpci_int=tpci.to_knx(),
                    payload_length=0,
                ),
            )[:4]
            secured_apdu = bytes(apdu)
        elif scf.algorithm == SecurityAlgorithmIdentifier.CCM_ENCRYPTION:
            mac_cbc = calculate_message_authentication_code_cbc(
                key=key,
                additional_data=scf.to_knx(),
                payload=apdu,
                block_0=block_0(
                    sequence_number=sequence_number_bytes,
                    address_fields_raw=address_fields_raw,
                    frame_flags=frame_flags,
                    tpci_int=tpci.to_knx(),
                    payload_length=len(apdu),
                ),
            )[:4]
            secured_apdu, mac = encrypt_data_ctr(
                key=key,
                counter_0=counter_0(
                    sequence_number=sequence_number_bytes,
                    address_fields_raw=address_fields_raw,
                ),
                mac_cbc=mac_cbc,
                payload=apdu,
            )
        else:
            raise DataSecureError(f"Unknown secure algorithm {scf.algorithm}")

        return SecureData(
            sequence_number_bytes=sequence_number_bytes,
            secured_apdu=secured_apdu,
            message_authentication_code=mac,  # only 4 bytes are used
        )

    def to_knx(self) -> bytes:
        """Serialize to KNX raw data."""
        return (
            self.sequence_number_bytes
            + self.secured_apdu
            + self.message_authentication_code
        )

    @staticmethod
    def from_knx(raw: bytes) -> SecureData:
        """Parse/deserialize from KNX raw data."""
        return SecureData(
            sequence_number_bytes=raw[:6],
            secured_apdu=raw[6:-4],
            message_authentication_code=raw[-4:],
        )

    def get_plain_apdu(
        self,
        key: bytes,
        scf: SecurityControlField,
        address_fields_raw: bytes,
        frame_flags: int,
        tpci: TPCI,
    ) -> bytes:
        """
        Get plain APDU as raw bytes. Decrypted or verified depending on algorithm.

        Sequence number and sender individual address shall already be checked against
        Security Individual Address Table before calling this method.
        """
        if scf.algorithm == SecurityAlgorithmIdentifier.CCM_ENCRYPTION:
            dec_payload, mac_tr = decrypt_ctr(
                key=key,
                counter_0=counter_0(
                    sequence_number=self.sequence_number_bytes,
                    address_fields_raw=address_fields_raw,
                ),
                mac=self.message_authentication_code,
                payload=self.secured_apdu,
            )
            mac_cbc = calculate_message_authentication_code_cbc(
                key=key,
                additional_data=scf.to_knx(),
                payload=dec_payload,
                block_0=block_0(
                    sequence_number=self.sequence_number_bytes,
                    address_fields_raw=address_fields_raw,
                    frame_flags=frame_flags,
                    tpci_int=tpci.to_knx(),
                    payload_length=len(dec_payload),
                ),
            )[:4]
            if mac_cbc != mac_tr:
                raise DataSecureError("Data Secure MAC verification failed")
            return dec_payload

        if scf.algorithm == SecurityAlgorithmIdentifier.CCM_AUTHENTICATION:
            mac = calculate_message_authentication_code_cbc(
                key=key,
                additional_data=scf.to_knx() + self.secured_apdu,
                block_0=block_0(
                    sequence_number=self.sequence_number_bytes,
                    address_fields_raw=address_fields_raw,
                    frame_flags=frame_flags,
                    tpci_int=tpci.to_knx(),
                    payload_length=0,
                ),
            )[:4]
            if mac != self.message_authentication_code:
                raise DataSecureError("Data Secure MAC verification failed.")
            return self.secured_apdu

        raise DataSecureError(f"Unknown secure algorithm {scf.algorithm}")

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<SecureData "
            f"sequence_number={int.from_bytes(self.sequence_number_bytes, 'big')} "
            f'secured_apdu="{self.secured_apdu.hex()}" '
            f'message_authentication_code="{self.message_authentication_code.hex()}" />'
        )



"""Implementation of KNX XYY color data point type."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any

from .dpt import DPTComplex, DPTComplexData
from .payload import DPTArray, DPTBinary


@dataclass(slots=True)
class XYYColor(DPTComplexData):
    """
    Representation of XY color with brightness.

    `color`: tuple(x-axis, y-axis) each 0..1; None if invalid.
    `brightness`: int 0..255; None if invalid.
    """

    color: tuple[float, float] | None = None
    brightness: int | None = None

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> XYYColor:
        """Init from a dictionary."""
        color = None
        brightness = data.get("brightness")
        x_axis = data.get("x_axis")
        y_axis = data.get("y_axis")

        if x_axis is not None and y_axis is not None:
            try:
                color = (float(x_axis), float(y_axis))
            except ValueError as err:
                raise ValueError(f"Invalid value for color axis: {err}") from err
        elif x_axis is not None or y_axis is not None:
            raise ValueError("Both x_axis and y_axis must be provided")

        if brightness is not None:
            try:
                brightness = int(brightness)
            except ValueError as err:
                raise ValueError(f"Invalid value for brightness: {err}") from err

        return cls(color=color, brightness=brightness)

    def as_dict(self) -> dict[str, int | float | None]:
        """Create a JSON serializable dictionary."""
        return {
            "x_axis": self.color[0] if self.color is not None else None,
            "y_axis": self.color[1] if self.color is not None else None,
            "brightness": self.brightness,
        }

    def __or__(self, other: XYYColor) -> XYYColor:
        """Merge two XYYColor objects using only valid values."""
        return XYYColor(
            color=other.color if other.color is not None else self.color,
            brightness=other.brightness
            if other.brightness is not None
            else self.brightness,
        )


class DPTColorXYY(DPTComplex[XYYColor]):
    """Abstraction for KNX 6 octet color xyY (DPT 242.600)."""

    data_type = XYYColor
    payload_type = DPTArray
    payload_length = 6
    dpt_main_number = 242
    dpt_sub_number = 600
    value_type = "color_xyy"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> XYYColor:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)

        x_axis_int = raw[0] << 8 | raw[1]
        y_axis_int = raw[2] << 8 | raw[3]
        brightness = raw[4]

        color_valid = raw[5] >> 1 & 0b1
        brightness_valid = raw[5] & 0b1

        return XYYColor(
            color=(
                # round to 5 digits for better readability but still preserving precision
                round(x_axis_int / 0xFFFF, 5),
                round(y_axis_int / 0xFFFF, 5),
            )
            if color_valid
            else None,
            brightness=brightness if brightness_valid else None,
        )

    @classmethod
    def _to_knx(cls, value: XYYColor) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        x_axis, y_axis, brightness = 0, 0, 0
        color_valid = False
        brightness_valid = False

        if value.color is not None:
            for axis in value.color:
                if not 0 <= axis <= 1:
                    raise ValueError(
                        f"Color axis value out of range 0..1: {value.color}"
                    )
            x_axis, y_axis = (round(axis * 0xFFFF) for axis in value.color)
            color_valid = True

        if value.brightness is not None:
            brightness = value.brightness
            if not 0 <= brightness <= 255:
                raise ValueError(f"Brightness out of range 0..255: {brightness}")
            brightness_valid = True

        return DPTArray(
            (
                x_axis >> 8,
                x_axis & 0xFF,
                y_axis >> 8,
                y_axis & 0xFF,
                brightness,
                color_valid << 1 | brightness_valid,
            )
        )



"""Unit test for KNX 2 byte float objects."""

import pytest

from xknx.dpt import (
    DPT2ByteFloat,
    DPTAirFlow,
    DPTArray,
    DPTEnthalpy,
    DPTHumidity,
    DPTLux,
    DPTPartsPerMillion,
    DPTTemperature,
    DPTVoltage,
)
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPTFloat:
    """Test class for KNX 2 byte/octet float object."""

    # ####################################################################
    # DPT2ByteFloat
    #
    def test_value_from_documentation(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat -30.00. Example from the internet[tm]."""
        assert DPT2ByteFloat.to_knx(-30.00) == DPTArray((0x8A, 0x24))
        assert DPT2ByteFloat.from_knx(DPTArray((0x8A, 0x24))) == -30.00

    def test_value_taken_from_live_thermostat(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat 19.96."""
        assert DPT2ByteFloat.to_knx(16.96) == DPTArray((0x06, 0xA0))
        assert DPT2ByteFloat.from_knx(DPTArray((0x06, 0xA0))) == 16.96

    def test_zero_value(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat zero value."""
        assert DPT2ByteFloat.to_knx(0.00) == DPTArray((0x00, 0x00))
        assert DPT2ByteFloat.from_knx(DPTArray((0x00, 0x00))) == 0.00

    def test_near_zero_value(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat near zero value."""
        assert DPT2ByteFloat.to_knx(0.0002) == DPTArray((0x00, 0x00))
        assert DPT2ByteFloat.to_knx(0.005) == DPTArray((0x00, 0x00))
        assert DPT2ByteFloat.to_knx(0.00501) == DPTArray((0x00, 0x01))

        assert DPT2ByteFloat.to_knx(-0.0) == DPTArray((0x00, 0x00))
        # ETS would convert values < 0 and >= -0.005 to 0x8000
        # which is equivalent to -20.48 so we handle this differently
        assert DPT2ByteFloat.to_knx(-0.0002) == DPTArray((0x00, 0x00))
        assert DPT2ByteFloat.to_knx(-0.005) == DPTArray((0x00, 0x00))
        assert DPT2ByteFloat.to_knx(-0.00501) == DPTArray(
            (0x87, 0xFF)
        )  # this is ETS-conform again

        assert DPT2ByteFloat.from_knx(DPTArray((0x00, 0x01))) == 0.01
        assert DPT2ByteFloat.from_knx(DPTArray((0x87, 0xFF))) == -0.01

    def test_room_temperature(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat 21.00. Room temperature."""
        assert DPT2ByteFloat.to_knx(21.00) == DPTArray((0x0C, 0x1A))
        assert DPT2ByteFloat.from_knx(DPTArray((0x0C, 0x1A))) == 21.00

    def test_high_temperature(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat 500.00, 499.84, 500.16. Testing rounding issues."""
        assert DPT2ByteFloat.to_knx(500.00) == DPTArray((0x2E, 0x1A))
        assert (
            round(abs(DPT2ByteFloat.from_knx(DPTArray((0x2E, 0x1A))) - 499.84), 7) == 0
        )
        assert (
            round(abs(DPT2ByteFloat.from_knx(DPTArray((0x2E, 0x1B))) - 500.16), 7) == 0
        )
        assert DPT2ByteFloat.to_knx(499.84) == DPTArray((0x2E, 0x1A))
        assert DPT2ByteFloat.to_knx(500.16) == DPTArray((0x2E, 0x1B))

    def test_minor_negative_temperature(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat -10.00. Testing negative values."""
        assert DPT2ByteFloat.to_knx(-10.00) == DPTArray((0x84, 0x18))
        assert DPT2ByteFloat.from_knx(DPTArray((0x84, 0x18))) == -10.00

    def test_very_cold_temperature(self) -> None:
        """
        Test parsing and streaming of DPT2ByteFloat -1000.00,-999.68, -1000.32.

        Testing rounding issues of negative values.
        """
        assert DPT2ByteFloat.to_knx(-1000.00) == DPTArray((0xB1, 0xE6))
        assert DPT2ByteFloat.from_knx(DPTArray((0xB1, 0xE6))) == -999.68
        assert DPT2ByteFloat.from_knx(DPTArray((0xB1, 0xE5))) == -1000.32
        assert DPT2ByteFloat.to_knx(-999.68) == DPTArray((0xB1, 0xE6))
        assert DPT2ByteFloat.to_knx(-1000.32) == DPTArray((0xB1, 0xE5))

    def test_max(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat with maximum value."""
        assert DPT2ByteFloat.to_knx(DPT2ByteFloat.value_max) == DPTArray((0x7F, 0xFF))
        assert DPT2ByteFloat.from_knx(DPTArray((0x7F, 0xFF))) == DPT2ByteFloat.value_max

    def test_close_to_limit(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat with numeric limit."""
        assert DPT2ByteFloat.to_knx(20.48) == DPTArray((0x0C, 0x00))
        assert DPT2ByteFloat.from_knx(DPTArray((0x0C, 0x00))) == 20.48
        assert DPT2ByteFloat.to_knx(-20.48) == DPTArray((0x80, 0x00))
        assert DPT2ByteFloat.from_knx(DPTArray((0x80, 0x00))) == -20.48

    def test_min(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat with minimum value."""
        assert DPT2ByteFloat.to_knx(DPT2ByteFloat.value_min) == DPTArray((0xF8, 0x00))
        assert DPT2ByteFloat.from_knx(DPTArray((0xF8, 0x00))) == DPT2ByteFloat.value_min

    def test_close_to_max(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat with maximum value -1."""
        assert DPT2ByteFloat.to_knx(670433.28) == DPTArray((0x7F, 0xFE))
        assert DPT2ByteFloat.from_knx(DPTArray((0x7F, 0xFE))) == 670433.28

    def test_close_to_min(self) -> None:
        """Test parsing and streaming of DPT2ByteFloat with minimum value +1."""
        assert DPT2ByteFloat.to_knx(-670760.96) == DPTArray((0xF8, 0x01))
        assert DPT2ByteFloat.from_knx(DPTArray((0xF8, 0x01))) == -670760.96

    def test_to_knx_min_exceeded(self) -> None:
        """Test parsing of DPT2ByteFloat with wrong value (underflow)."""
        with pytest.raises(ConversionError):
            DPT2ByteFloat.to_knx(DPT2ByteFloat.value_min - 1)

    def test_to_knx_max_exceeded(self) -> None:
        """Test parsing of DPT2ByteFloat with wrong value (overflow)."""
        with pytest.raises(ConversionError):
            DPT2ByteFloat.to_knx(DPT2ByteFloat.value_max + 1)

    def test_to_knx_wrong_parameter(self) -> None:
        """Test parsing of DPT2ByteFloat with wrong value (string)."""
        with pytest.raises(ConversionError):
            DPT2ByteFloat.to_knx("fnord")

    def test_from_knx_wrong_parameter(self) -> None:
        """Test parsing of DPT2ByteFloat with wrong value (wrong number of bytes)."""
        with pytest.raises(CouldNotParseTelegram):
            DPT2ByteFloat.from_knx(DPTArray((0xF8, 0x01, 0x23)))

    #
    # DPTTemperature
    #
    def test_temperature_settings(self) -> None:
        """Test attributes of DPTTemperature."""
        assert DPTTemperature.value_min == -273
        assert DPTTemperature.value_max == 670760
        assert DPTTemperature.unit == "°C"
        assert DPTTemperature.resolution == 0.01

    def test_temperature_assert_min_exceeded(self) -> None:
        """Testing parsing of DPTTemperature with wrong value."""
        with pytest.raises(ConversionError):
            DPTTemperature.to_knx(-274)

    def test_temperature_assert_min_exceeded_from_knx(self) -> None:
        """Testing parsing of DPTTemperature with wrong value."""
        with pytest.raises(ConversionError):
            DPTTemperature.from_knx(DPTArray((0xB1, 0xE6)))  # -1000

    #
    # DPTLux
    #
    def test_lux_settings(self) -> None:
        """Test attributes of DPTLux."""
        assert DPTLux.value_min == 0
        assert DPTLux.value_max == 670760
        assert DPTLux.unit == "lx"
        assert DPTLux.resolution == 0.01

    def test_lux_assert_min_exceeded(self) -> None:
        """Test parsing of DPTLux with wrong value."""
        with pytest.raises(ConversionError):
            DPTLux.to_knx(-1)

    #
    # DPTHumidity
    #
    def test_humidity_settings(self) -> None:
        """Test attributes of DPTHumidity."""
        assert DPTHumidity.value_min == 0
        assert DPTHumidity.value_max == 670760
        assert DPTHumidity.unit == "%"
        assert DPTHumidity.resolution == 0.01

    def test_humidity_assert_min_exceeded(self) -> None:
        """Test parsing of DPTHumidity with wrong value."""
        with pytest.raises(ConversionError):
            DPTHumidity.to_knx(-1)

    #
    # DPTEnthalpy
    #
    def test_enthalpy_settings(self) -> None:
        """Test attributes of DPTEnthalpy."""
        assert DPTEnthalpy.unit == "H"

    #
    # DPTPartsPerMillion
    #
    def test_partspermillion_settings(self) -> None:
        """Test attributes of DPTPartsPerMillion."""
        assert DPTPartsPerMillion.unit == "ppm"

    #
    # DPTAirFlow
    #
    def test_airflow_settings(self) -> None:
        """Test attributes of DPTAirFlow."""
        assert DPTAirFlow.unit == "m³/h"

    #
    # DPTVoltage
    #
    def test_voltage_settings(self) -> None:
        """Test attributes of DPTVoltage."""
        assert DPTVoltage.unit == "mV"



"""Implementation of Basic KNX 4-Byte unsigned integer values."""

from __future__ import annotations

from .dpt import DPTNumeric, DPTStructIntMixin


class DPT4ByteUnsigned(DPTStructIntMixin, DPTNumeric):
    """
    Abstraction for KNX 4 Byte "32-bit unsigned".

    DPT 12.***
    """

    dpt_main_number = 12
    dpt_sub_number: int | None = None
    value_type = "4byte_unsigned"
    payload_length = 4

    value_min = 0
    value_max = 4294967295
    resolution = 1

    _struct_format = ">I"


class DPTValue4Ucount(DPT4ByteUnsigned):
    """DPT 12.001 DPT_Value_4_Ucount."""

    dpt_main_number = 12
    dpt_sub_number = 1
    value_type = "pulse_4_ucount"
    unit = "counter pulses"


class DPTLongTimePeriodSec(DPT4ByteUnsigned):
    """DPT 12.100 DPT_LongTimePeriod_Sec (seconds)."""

    dpt_main_number = 12
    dpt_sub_number = 100
    value_type = "long_time_period_sec"
    unit = "s"


class DPTLongTimePeriodMin(DPT4ByteUnsigned):
    """DPT 12.101 DPT_LongTimePeriod_Min (minutes)."""

    dpt_main_number = 12
    dpt_sub_number = 101
    value_type = "long_time_period_min"
    unit = "min"


class DPTLongTimePeriodHrs(DPT4ByteUnsigned):
    """DPT 12.102 DPT_LongTimePeriod_Hrs (hours)."""

    dpt_main_number = 12
    dpt_sub_number = 102
    value_type = "long_time_period_hrs"
    unit = "h"


class DPTVolumeLiquidLitre(DPT4ByteUnsigned):
    """DPT 12.1200 DPT_VolumeLiquid_Litre (water/gas total consumption)."""

    dpt_main_number = 12
    dpt_sub_number = 1200
    value_type = "volume_liquid_litre"
    unit = "L"


class DPTVolumeM3(DPT4ByteUnsigned):
    """DPT 12.1201 DPT_Volume_m3 (water/gas total consumption volume)."""

    dpt_main_number = 12
    dpt_sub_number = 1201
    value_type = "volume_m3"
    unit = "m³"



"""Unit test for KNX/IP SRP objects."""

import pytest

from xknx.exceptions import ConversionError, CouldNotParseKNXIP
from xknx.knxip import DIBServiceFamily, DIBTypeCode, SearchRequestParameterType
from xknx.knxip.srp import SRP


class TestKNXIPSRP:
    """Test class for KNX/IP SRP objects."""

    def test_basic(self) -> None:
        """Test SRP with mac contains mac address."""
        srp: SRP = SRP.with_mac_address(bytes([1, 2, 3, 4, 5, 6]))
        assert len(srp) == SRP.SRP_HEADER_SIZE + 6
        assert srp.type == SearchRequestParameterType.SELECT_BY_MAC_ADDRESS

        srp: SRP = SRP.with_service(DIBServiceFamily.TUNNELING, 2)
        assert len(srp) == SRP.SRP_HEADER_SIZE + 2
        assert srp.type == SearchRequestParameterType.SELECT_BY_SERVICE

        srp: SRP = SRP.with_programming_mode()
        assert len(srp) == SRP.SRP_HEADER_SIZE
        assert srp.type == SearchRequestParameterType.SELECT_BY_PROGRAMMING_MODE

        dibs: list[DIBTypeCode] = [
            DIBTypeCode.SUPP_SVC_FAMILIES,
            DIBTypeCode.ADDITIONAL_DEVICE_INFO,
        ]
        srp: SRP = SRP.request_device_description(dibs)
        assert len(srp) == SRP.SRP_HEADER_SIZE + len(dibs)
        assert srp.type == SearchRequestParameterType.REQUEST_DIBS

    def test_invalid_payload_raises(self) -> None:
        """Test SRP with invalid data size raises."""
        with pytest.raises(ConversionError):
            SRP.with_mac_address(bytes([1, 2, 3, 4, 5, 6, 7, 8]))

        with pytest.raises(ConversionError):
            SRP(SearchRequestParameterType.SELECT_BY_SERVICE, True)

        with pytest.raises(ConversionError):
            SRP(SearchRequestParameterType.REQUEST_DIBS, True)

    def test_invalid_payload_from_knx_raises(self) -> None:
        """Test from_knx with invalid data size raises."""
        # size is too big
        with pytest.raises(CouldNotParseKNXIP):
            SRP.from_knx(bytes.fromhex("08 82"))

        # too small
        with pytest.raises(CouldNotParseKNXIP):
            SRP.from_knx(bytes.fromhex("08"))

    @pytest.mark.parametrize(
        ("srp", "expected_bytes"),
        [
            (
                SRP.with_mac_address(bytes([1, 2, 3, 4, 5, 6])),
                bytes.fromhex("08 82 01 02 03 04 05 06"),
            ),
            (
                SRP.with_service(DIBServiceFamily.TUNNELING, 2),
                bytes.fromhex("04 83 04 02"),
            ),
            (
                SRP.with_programming_mode(),
                bytes.fromhex("02 81"),
            ),
            (
                SRP.request_device_description(
                    [DIBTypeCode.SUPP_SVC_FAMILIES, DIBTypeCode.TUNNELING_INFO]
                ),
                bytes.fromhex("04 04 02 07"),
            ),
            (
                SRP.request_device_description([DIBTypeCode.SUPP_SVC_FAMILIES]),
                bytes.fromhex("04 04 02 00"),
            ),
        ],
    )
    def test_to_and_from_knx(self, srp: SRP, expected_bytes: bytes) -> None:
        """Test to and from KNX."""
        assert bytes(srp) == expected_bytes
        assert SRP.from_knx(expected_bytes) == srp



"""Unit test for Scene objects."""

from unittest.mock import Mock

from xknx import XKNX
from xknx.devices import Scene
from xknx.dpt import DPTArray
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestScene:
    """Test class for Scene objects."""

    #
    # SYNC
    #
    async def test_sync(self) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        scene = Scene(xknx, "TestScene", group_address="1/2/1", scene_number=23)
        await scene.sync()
        assert xknx.telegrams.qsize() == 0

    #
    # TEST RUN SCENE
    #
    async def test_run(self) -> None:
        """Test running scene."""
        xknx = XKNX()
        scene = Scene(xknx, "TestScene", group_address="1/2/1", scene_number=23)
        await scene.run()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/1"),
            payload=GroupValueWrite(DPTArray(0x16)),
        )

    #
    # TEST has_group_address
    #
    def test_has_group_address(self) -> None:
        """Test has_group_address."""
        xknx = XKNX()
        scene = Scene(xknx, "TestScene", group_address="1/2/1", scene_number=23)
        assert scene.has_group_address(GroupAddress("1/2/1"))
        assert not scene.has_group_address(GroupAddress("2/2/2"))

    async def test_process_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback is called."""

        xknx = XKNX()
        after_update_callback = Mock()
        scene = Scene(
            xknx,
            "TestScene",
            group_address="1/2/3",
            scene_number=1,
            device_updated_cb=after_update_callback,
        )

        scene.process(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(DPTArray((0x00,))),
            )
        )
        after_update_callback.assert_called_with(scene)

        after_update_callback.reset_mock()
        scene.process(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(DPTArray((0x01,))),  # different scene number
            )
        )
        after_update_callback.assert_not_called()



"""Unit test for KNX/IP ConnectResponses."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import (
    HPAI,
    ConnectRequestType,
    ConnectResponse,
    ConnectResponseData,
    ErrorCode,
    KNXIPFrame,
)
from xknx.telegram import IndividualAddress


class TestKNXIPConnectResponse:
    """Test class for KNX/IP ConnectResponses."""

    def test_tunnel_connect_response(self) -> None:
        """Test parsing and streaming connection response KNX/IP packet."""
        raw = bytes.fromhex(
            "06 10 02 06 00 14 01 00 08 01 C0 A8 2A 0A 0E 57 04 04 11 FF"
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)
        assert isinstance(knxipframe.body, ConnectResponse)
        assert knxipframe.body.communication_channel == 1
        assert knxipframe.body.status_code == ErrorCode.E_NO_ERROR
        assert knxipframe.body.data_endpoint == HPAI(ip_addr="192.168.42.10", port=3671)
        assert knxipframe.body.crd.request_type == ConnectRequestType.TUNNEL_CONNECTION
        assert knxipframe.body.crd.individual_address.raw == 4607

        connect_response = ConnectResponse(
            communication_channel=1,
            status_code=ErrorCode.E_NO_ERROR,
            data_endpoint=HPAI(ip_addr="192.168.42.10", port=3671),
            crd=ConnectResponseData(
                request_type=ConnectRequestType.TUNNEL_CONNECTION,
                individual_address=IndividualAddress(4607),
            ),
        )
        knxipframe2 = KNXIPFrame.init_from_body(connect_response)

        assert knxipframe2.to_knx() == raw

    def test_mgmt_connect_response(self) -> None:
        """Test parsing and streaming connection response KNX/IP packet."""
        raw = bytes.fromhex("06 10 02 06 00 12 01 00 08 01 C0 A8 2A 0A 0E 57 02 03")
        knxipframe, _ = KNXIPFrame.from_knx(raw)
        assert isinstance(knxipframe.body, ConnectResponse)
        assert knxipframe.body.communication_channel == 1
        assert knxipframe.body.status_code == ErrorCode.E_NO_ERROR
        assert knxipframe.body.data_endpoint == HPAI(ip_addr="192.168.42.10", port=3671)
        assert (
            knxipframe.body.crd.request_type
            == ConnectRequestType.DEVICE_MGMT_CONNECTION
        )
        assert knxipframe.body.crd.individual_address is None

        connect_response = ConnectResponse(
            communication_channel=1,
            status_code=ErrorCode.E_NO_ERROR,
            data_endpoint=HPAI(ip_addr="192.168.42.10", port=3671),
            crd=ConnectResponseData(
                request_type=ConnectRequestType.DEVICE_MGMT_CONNECTION,
            ),
        )
        knxipframe2 = KNXIPFrame.init_from_body(connect_response)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_crd(self) -> None:
        """Test parsing and streaming wrong ConnectRequest (wrong CRD length byte)."""
        raw = bytes.fromhex(
            "06 10 02 06 00 14 01 00 08 01 C0 A8 2A 0A 0E 57 03 04 11 FF"
        )
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_crd2(self) -> None:
        """Test parsing and streaming wrong ConnectRequest (wrong CRD length)."""
        raw = bytes.fromhex("06 10 02 06 00 12 01 00 08 01 C0 A8 2A 0A 0E 57 04 04")
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_crd3(self) -> None:
        """Test parsing and streaming wrong ConnectRequest (CRD length too small)."""
        raw = bytes.fromhex("06 10 02 06 00 12 01 00 08 01 C0 A8 2A 0A 0E 57 01 04")
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_crd4(self) -> None:
        """Test parsing and streaming wrong ConnectRequest (wrong CRD length)."""
        raw = bytes.fromhex("06 10 02 06 00 12 01 00 08 01 C0 A8 2A 0A 0E 57 04 03")
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_crd5(self) -> None:
        """Test parsing and streaming wrong ConnectRequest (wrong CRD length)."""
        raw = bytes.fromhex("06 10 02 06 00 13 01 00 08 01 C0 A8 2A 0A 0E 57 03 03 01")
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_connect_response_connection_error_gira(self) -> None:
        """
        Test parsing and streaming connection response KNX/IP packet with error e_no_more_connections.

        HPAI and CRD normal. This was received from Gira devices (2020).
        """
        raw = bytes.fromhex(
            "06 10 02 06 00 14 C0 24 08 01 0A 01 00 29 0E 57 04 04 00 00"
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)
        assert isinstance(knxipframe.body, ConnectResponse)
        assert knxipframe.body.status_code == ErrorCode.E_NO_MORE_CONNECTIONS
        assert knxipframe.body.communication_channel == 192

        connect_response = ConnectResponse(
            communication_channel=192,
            status_code=ErrorCode.E_NO_MORE_CONNECTIONS,
            data_endpoint=HPAI(ip_addr="10.1.0.41", port=3671),
            crd=ConnectResponseData(
                request_type=ConnectRequestType.TUNNEL_CONNECTION,
                individual_address=IndividualAddress(0),
            ),
        )
        knxipframe2 = KNXIPFrame.init_from_body(connect_response)

        assert knxipframe2.to_knx() == raw

    def test_connect_response_connection_error_lox(self) -> None:
        """
        Test parsing and streaming connection response KNX/IP packet with error e_no_more_connections.

        HPAI given, CRD all zero. This was received from Loxone device (2020).
        """
        raw = bytes.fromhex(
            "06 10 02 06 00 14 00 24 08 01 C0 A8 01 01 0E 57 00 00 00 00"
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)
        assert isinstance(knxipframe.body, ConnectResponse)
        assert knxipframe.body.status_code == ErrorCode.E_NO_MORE_CONNECTIONS
        assert knxipframe.body.communication_channel == 0

        # no further tests: the current API can't (and shouldn't) create such odd packets

    def test_connect_response_connection_error_mdt(self) -> None:
        """
        Test parsing and streaming connection response KNX/IP packet with error e_no_more_connections.

        HPAI and CRD all zero. This was received from MDT device (2020).
        """
        raw = bytes.fromhex("06 10 02 06 00 08 00 24 00 00 00 00 00 00 00 00 00 00")
        knxipframe, _ = KNXIPFrame.from_knx(raw)
        assert isinstance(knxipframe.body, ConnectResponse)
        assert knxipframe.body.status_code == ErrorCode.E_NO_MORE_CONNECTIONS
        assert knxipframe.body.communication_channel == 0

        # no further tests: the current API can't (and shouldn't) create such odd packets



"""
Module for Serialization and Deserialization of KNX Description Response.

The DESCRIPTION_RESPONSE frame shall be sent by the KNXnet/IP Server as an answer to
a received DESCRIPTION_REQUEST frame. It shall be addressed to the KNXnet/IP Clients
control endpoint using the HPAI included in the received DESCRIPTION_REQUEST frame.
The size of the KNXnet/IP body varies depending on the number of DIB structures sent
by the KNXnet/IP Server in response to the KNXnet/IP Clients DESCRIPTION_REQUEST.
"""

from __future__ import annotations

from .body import KNXIPBody
from .dib import DIB, DIBDeviceInformation
from .knxip_enum import KNXIPServiceType


class DescriptionResponse(KNXIPBody):
    """Representation of a KNX Description Response."""

    SERVICE_TYPE = KNXIPServiceType.DESCRIPTION_RESPONSE

    def __init__(self) -> None:
        """Initialize SearchResponse object."""
        self.dibs: list[DIB] = []

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return sum(dib.calculated_length() for dib in self.dibs)

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        pos = 0
        while raw[pos:]:
            dib = DIB.determine_dib(raw[pos:])
            pos += dib.from_knx(raw[pos:])
            self.dibs.append(dib)
        return pos

    @property
    def device_name(self) -> str:
        """Return name of device."""
        return next(
            (dib.name for dib in self.dibs if isinstance(dib, DIBDeviceInformation)),
            "UNKNOWN",
        )

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return b"".join(dib.to_knx() for dib in self.dibs)

    def __repr__(self) -> str:
        """Return object as readable string."""
        _dibs_str = ",\n".join(dib.__repr__() for dib in self.dibs)
        return f'<DescriptionResponse dibs="[\n{_dibs_str}\n]" />'



"""Unit test for RemoteValueColorXYY objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary, XYYColor
from xknx.exceptions import ConversionError
from xknx.remote_value import RemoteValueColorXYY
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestRemoteValueColorXYY:
    """Test class for RemoteValueColorXYY objects."""

    def test_to_knx(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueColorXYY(xknx)
        assert remote_value.to_knx(XYYColor((1, 0.9), 102)) == DPTArray(
            (0xFF, 0xFF, 0xE6, 0x66, 0x66, 0x03)
        )
        assert remote_value.to_knx(XYYColor((1, 0), 102)) == DPTArray(
            (0xFF, 0xFF, 0x00, 0x00, 0x66, 0x03)
        )

    def test_from_knx(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueColorXYY(xknx)
        assert remote_value.from_knx(
            DPTArray((0x99, 0x99, 0x99, 0x99, 0x66, 0x03))
        ) == XYYColor((0.6, 0.6), 102)

    def test_to_knx_error(self) -> None:
        """Test to_knx function with wrong parametern."""
        xknx = XKNX()
        remote_value = RemoteValueColorXYY(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx(XYYColor((2, 1), 1))
        with pytest.raises(ConversionError):
            remote_value.to_knx(XYYColor((-1, 1), 2))
        with pytest.raises(ConversionError):
            remote_value.to_knx(XYYColor((0.3, 0.5), 256))
        with pytest.raises(ConversionError):
            remote_value.to_knx(XYYColor(("0.4", 0), 102))
        with pytest.raises(ConversionError):
            remote_value.to_knx(XYYColor((1, 1), "102"))
        with pytest.raises(ConversionError):
            remote_value.to_knx(XYYColor((1,), 1))

    async def test_set(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueColorXYY(xknx, group_address=GroupAddress("1/2/3"))
        remote_value.set(XYYColor((1, 0.9), 102))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0xFF, 0xFF, 0xE6, 0x66, 0x66, 0x03))),
        )
        remote_value.set(XYYColor((1, 0.9), 255))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0xFF, 0xFF, 0xE6, 0x66, 0xFF, 0x03))),
        )

    def test_process(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueColorXYY(xknx, group_address=GroupAddress("1/2/3"))
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0xFF, 0xFF, 0x66, 0x66, 0xFA, 0x03))),
        )
        remote_value.process(telegram)
        assert remote_value.value == XYYColor((1, 0.4), 250)

    def test_to_process_error(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueColorXYY(xknx, group_address=GroupAddress("1/2/3"))

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x66, 0x67))),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None



"""Implementation of Basic KNX DPT_1_Ucount Values."""

from __future__ import annotations

from xknx.exceptions import ConversionError

from .dpt import DPTNumeric
from .payload import DPTArray, DPTBinary


class DPTValue1ByteUnsigned(DPTNumeric):
    """
    Abstraction for KNX 1 Octet.

    DPT 5.***
    """

    dpt_main_number = 5
    dpt_sub_number: int | None = None
    value_type = "1byte_unsigned"
    payload_length = 1

    value_min = 0
    value_max = 255
    resolution = 1

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        value = cls.validate_payload(payload)[0]

        if not cls._test_boundaries(value):
            raise ConversionError(
                f"Could not parse {cls.dpt_name()}", value=value, payload=payload
            )

        return value

    @classmethod
    def to_knx(cls, value: int | float) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        try:
            knx_value = int(value)
            if not cls._test_boundaries(knx_value):
                raise ValueError(f"Value out of range {cls.value_min}..{cls.value_max}")
            return DPTArray(knx_value)
        except ValueError as err:
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}", value=value
            ) from err

    @classmethod
    def _test_boundaries(cls, value: int) -> bool:
        """Test if value is within defined range for this object."""
        return cls.value_min <= value <= cls.value_max


class DPTScaling(DPTNumeric):
    """
    Abstraction for KNX 1 Octet Percent.

    DPT 5.001
    """

    dpt_main_number = 5
    dpt_sub_number = 1
    value_type = "percent"
    unit = "%"
    payload_length = 1

    value_min = 0
    value_max = 100
    resolution = 1

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        knx_value = cls.validate_payload(payload)[0]
        delta = cls.value_max - cls.value_min
        value = round((knx_value / 255) * delta) + cls.value_min

        if not cls._test_boundaries(value):
            raise ConversionError(
                f"Could not parse {cls.dpt_name()}", value=value, payload=payload
            )
        return value

    @classmethod
    def to_knx(cls, value: float) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        try:
            percent_value = float(value)
            if not cls._test_boundaries(percent_value):
                raise ValueError("Value out of range")
            delta = cls.value_max - cls.value_min
            knx_value = round((percent_value - cls.value_min) / delta * 255)

            return DPTArray(knx_value)
        except ValueError as err:
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}", value=value
            ) from err

    @classmethod
    def _test_boundaries(cls, value: float) -> bool:
        """Test if value is within defined range for this object."""
        return cls.value_min <= value <= cls.value_max


class DPTAngle(DPTScaling):
    """
    Abstraction for KNX 1 Octet Angle.

    DPT 5.003
    """

    dpt_main_number = 5
    dpt_sub_number = 3
    value_type = "angle"
    unit = "°"

    value_min = 0
    value_max = 360
    resolution = 1


class DPTPercentU8(DPTValue1ByteUnsigned):
    """
    Abstraction for KNX 1 Octet Percent.

    DPT 5.004
    """

    dpt_main_number = 5
    dpt_sub_number = 4
    value_type = "percentU8"
    unit = "%"
    resolution = 1


class DPTDecimalFactor(DPTValue1ByteUnsigned):
    """
    Abstraction for KNX 1 Octet Percent.

    DPT 5.005
    """

    dpt_main_number = 5
    dpt_sub_number = 5
    value_type = "decimal_factor"


class DPTTariff(DPTValue1ByteUnsigned):
    """
    Abstraction for KNX 1 Octet tariff information.

    DPT 5.006
    """

    dpt_main_number = 5
    dpt_sub_number = 6
    value_type = "tariff"
    value_max = 254


class DPTValue1Ucount(DPTValue1ByteUnsigned):
    """
    Abstraction for KNX 1 Octet counter pulses.

    DPT 5.010
    """

    dpt_main_number = 5
    dpt_sub_number = 10
    value_type = "pulse"
    unit = "counter pulses"
